import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;

/** This class is used to handle the borders.
 * It includes functions for adjusting the thickness, color,
 * and even printing the borders.
 *@author TSS
 * @version 1.0 03/09/99
 */
public class LinesBorder extends AbstractBorder implements SwingConstants { 
 
  
  /** The thickness of the north border.*/
	   protected int northThickness;
	   /** The thickness of the south border. */
	   protected int southThickness;
	   /** The thickness of the east border. */
	   protected int eastThickness;
	   /** The thickness of the west border. */
	   protected int westThickness;  
	   /** The color of the north border. */
	   protected Color northColor;
	   /** The color of the south border. */
	   protected Color southColor;
	   /** The color of the east border. */
	   protected Color eastColor;
	   /** The color of the west border. */
	protected Color westColor;
  
  /** This constructor sets the initial border color from
   * the parameter and defaults the thickness to 1.
   * @param color  - an Color object.
   */  
  public LinesBorder(Color color) {
    this(color, 1);
  }

  /** This constructor sets the initial color and thickness of the
   * borders.
   * @param color  - a Color object.
   * @param thickness - an integer value for thickness for all boarders.
   */  
  public LinesBorder(Color color, int thickness)  {
    setColor(color);
    setThickness(thickness);
  }

  /** This constructor sets the initial color and
   * individual thicknesses for each border.
   * @param color  - a Color object.
   * @param insets  - an Insets object with all four thicknesses.
   */  
  public LinesBorder(Color color, Insets insets)  {
    setColor(color);
    setThickness(insets);
  }

  /** This function paints the borders on the passed-in graphics.
   * This is used to show the border graphically for printing
   * purposes.
   * @param c  - a Component object.
   * @param g - a Graphics object used to draw the lines on.
   * @param x  - an integer for the x value.
   * @param y  - an integer for the y value.
   * @param width - an integer for the width of the whole border box.
   * @param height - an integer for the height of the whole border box.
   */  
  public void paintBorder(Component c, Graphics g, int x, int y, int width, int height) {
    Color oldColor = g.getColor();
    
    g.setColor(Color.black);
    for (int i = 0; i < northThickness; i++)  {
      g.drawLine(x, y+i, x+width-1, y+i);
    }
    //g.setColor(Color.black);
    for (int i = 0; i < southThickness; i++)  {
      g.drawLine(x, y+height-i-1, x+width-1, y+height-i-1);
    }
    //g.setColor(Color.black);
    for (int i = 0; i < westThickness; i++)  {
      g.drawLine(x+i, y, x+i, y+height-1);
    }
    //g.setColor(Color.black);
    for (int i = 0; i < eastThickness; i++)  {
      g.drawLine(x+width-i-1, y, x+width-i-1, y+height-1);
    }
 
    g.setColor(oldColor);
  }

  /** This function gets Border attributes in an Insets
   * object.
   * @param c  - a Component object (Not used in function)
   * @return  - an Insets object with all border thicknesses.
   */  
  public Insets getBorderInsets(Component c)       {
    return new Insets(northThickness, westThickness, southThickness, eastThickness);
  }
 
  /** This function gets the Border attributes as an
   * Inset object.
   * @param c  - a Component Object (Not used in function)
   * @param insets  - an Insets object (Not used in function)
   * @return  - an Insets object with all thicknesses.
   */  
  public Insets getBorderInsets(Component c, Insets insets) {
    return new Insets(northThickness, westThickness, southThickness, eastThickness);    
  }
 
  /** This function returns whether the borders
   * are Opaque or not.
   * Set to always return TRUE.
   * @return  - returns TRUE always.
   */  
  public boolean isBorderOpaque() { return true; }
    
  /** This function sets the same color for all
   * four borders.
   * @param c  - a Color object.
   */  
  public void setColor(Color c) {
    northColor = new Color (c.getRGB());
    southColor = new Color (c.getRGB());
    eastColor  = new Color (c.getRGB());
    westColor  = new Color (c.getRGB());
  }
  
  /** This function sets the color for that particular
   * border.
   * @param c  - a Color object.
   *
   * @param direction  - an integer value for the direction.
   */  
  public void setColor(Color c, int direction) {
    switch (direction) {
      case NORTH: northColor = new Color(c.getRGB()); break;
      case SOUTH: southColor = new Color(c.getRGB()); break;
      case EAST:  eastColor  = new Color(c.getRGB()); break;
      case WEST:  westColor  = new Color(c.getRGB()); break;
      default: 
    }
  }

  /** This function gets the color of the border in
   * the given direction.
   * The if invalid direction is passed in,
   * the default color is white.
   * @param direction  - an integer representing the direction.
   * @return  - a Color object.
   */  
  public Color getColor(int direction) {
    switch (direction) {
      case NORTH: return northColor;
      case SOUTH: return southColor;
      case EAST:  return eastColor;
      case WEST:  return westColor;
    }
    return Color.white;
  }
    
  /** This function sets the thickness for all four borders.
   * @param n  - an integer value for the thickness.
   */  
  public void setThickness(int n) {
    northThickness = n;
    southThickness = n;
    eastThickness  = n;
    westThickness  = n;
  }
    
  /** This function takes in an Inset object which contains
   * the thickness for all four dirctions.
   * @param insets  - an Inset object with all four thicknesses.
   */  
  public void setThickness(Insets insets) {
    northThickness = insets.top;
    southThickness = insets.bottom;
    eastThickness  = insets.right;
    westThickness  = insets.left;
  }
  
  /** This function sets the thickness for the given
   * side.
   * @param n  - an integer for the thickness.
   * @param direction  - an integer value for the direction.
   */  
  public void setThickness(int n, int direction) {
    switch (direction) {
      case NORTH: northThickness = n; break;
      case SOUTH: southThickness = n; break;
      case EAST:  eastThickness  = n; break;
      case WEST:  westThickness  = n; break;
      default: 
    }
  }

  /** This function gets the thickness for the given direction.
   * @param direction  - an integer representing the directions.
   *
   * @return  - an integer representing the thickness of the
   *   border on the given side.
   */  
  public int getThickness(int direction) {
    switch (direction) {
      case NORTH: return northThickness;
      case SOUTH: return southThickness;
      case EAST:  return eastThickness;
      case WEST:  return westThickness;
      default: 
    }
    return 0;
  } 
 
  /** This function adjusts the thickness of borders using
   * the parameters.
   * The function works similar to append(Inset, isReplace)
   * function.
   * @param b  -  a LinesBorder object
   * @param isReplace  - a boolean value.
   */  
  public void append(LinesBorder b, boolean isReplace) {
    if (isReplace) {
      northThickness = b.northThickness;
      southThickness = b.southThickness;
      eastThickness  = b.eastThickness;
      westThickness  = b.westThickness;
    } else {
      northThickness = Math.max(northThickness ,b.northThickness);
      southThickness = Math.max(southThickness ,b.southThickness);
      eastThickness  = Math.max(eastThickness  ,b.eastThickness);
      westThickness  = Math.max(westThickness  ,b.westThickness);
    }
  }

  /** This function adjusts the thickness of the borders
   * using the parameters.
   * If IsReplace == true then
   *      - the parameter insets is used to replace the
   *      existing border thickness.
   * If IsReplace == false then
   *       - the function finds the max of the inset thickness
   *      and the existing border thickness and uses the
   *      max of the two for each side.
   * @param insets  - an Inset object with border thickness.
   * @param isReplace  - a boolean value.
   */  
  public void append(Insets insets, boolean isReplace) {
    if (isReplace) {
      northThickness = insets.top;
      southThickness = insets.bottom;
      eastThickness  = insets.right;
      westThickness  = insets.left;
    } else {
      northThickness = Math.max(northThickness ,insets.top);
      southThickness = Math.max(southThickness ,insets.bottom);
      eastThickness  = Math.max(eastThickness  ,insets.right);
      westThickness  = Math.max(westThickness  ,insets.left);
    }
  }
}



