import java.awt.Color;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.Insets;
import java.io.BufferedReader;
import java.io.StringReader;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JTable;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;


//**************************************************************
//Class: UMTableModel --
//Description: this is the main backend class for the spreadsheet
//             it contains the model of our table.
//Disclaimer: Some of the code and and ideas we found online in
//            the ShartTools project.  We thank them for making
//            their code availible.
//*************************************************************

/** This is the main backend class for the spreadsheet. It contains the model of our table. 
 * Each aspect of the TerpSpreadSheet (such as setting or retrieving values and fonts)
 * is manipulated and referenced through this class.
 * @author TSS
 * @version 3.0
 */
public class UMTableModel extends DefaultTableModel {
    // Stores modified state of document
	/**
		* Stores modified state of document
		*/
    private boolean m_modified;
    
    
    /**  Stores file name of current document */
    private TerpSpreadSheet m_spreadsheet;
    
	/**
	  *  The list of all available fonts
	  */
    private String Fontlist[]=GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
    /** Default Constructor. 
     * <P>Sets the reference for future functions to the current spreadsheet by associating the
     * passed TerpSpreadSheet with the current UMTableModel.
     * @param spreadsheet  The current spreadsheet object.*/ 
    public UMTableModel(TerpSpreadSheet spreadsheet) {
        super();
        
        // initialize state to unmodified and file to untitled
        m_modified = false;
        this.m_spreadsheet = spreadsheet;
    }
    
	/**
	   * Default constructor
	   */
	public UMTableModel()
	{
		super();
		m_modified = false;
	}
        //----------------------------------
    
    
    /** Constructor. Creates an empty table with numRow rows and numCol columns 
     * plus the header row and column. <P>The zeroth row and column are headers. 
     * Rows are labeled 1, 2, etc and columns are labeled A, B, etc. The table is
     * initialized with no cell objects. Each coordinate that will have data will 
     * have a cell object created later.
     * @param spreadsheet The spreadsheet to be created and associated with this UMTableModel.
     * @param numRows Number of rows to create (including row header). Must be a positive integer.
     * @param numColumns  Number of columns to create (including col header). Must be a positive integer. */    
    public UMTableModel(TerpSpreadSheet spreadsheet,
    int numRows,
    int numColumns) {
        
        super(numRows, numColumns+1);
        //initialize row headers
        for(int row = TerpSpreadSheet.m_baseRow; row < numRows; row++) {
            UMCell temp = new UMCell(new Integer(row+1));
            super.setValueAt(temp, row, 0);
        }
        
        //initialize all the cells
        for(int row = TerpSpreadSheet.m_baseRow; row < numRows; row++)
            for (int col = TerpSpreadSheet.m_baseCol; col < numColumns+1; col++)
                // we initialize it here
                super.setValueAt(new UMCell(""), row, col);
        
        // initialize state to unmodified and file to untitled
        m_modified = false;
        this.m_spreadsheet = spreadsheet;
    }
    
    /** Creates a new spreadsheet and loads it with values from the passed array.
     * @param spreadsheet The spreadsheet to be created and associated with this UMTableModel.
     * @param data  Array of objects to initialize the spreadsheet with. */    
    public UMTableModel(TerpSpreadSheet spreadsheet, Object[][] data) {
        
        this(spreadsheet, spreadsheet.m_maxRow + 1,
        spreadsheet.m_maxCol + 1);
        
        /* load the data */
        for(int i = TerpSpreadSheet.m_baseRow; i < data.length &&
        i <= spreadsheet.m_maxRow; i++) {
            for(int j = TerpSpreadSheet.m_baseCol; j < data[i].length &&
            j <= spreadsheet.m_maxCol; j++) {
                doSetValueAt(data[i][j], i, j);
            }
        }
        
        // initialize state to unmodified and file to untitled
        m_modified = false;
        //	file = new File("Untitled");
        this.m_spreadsheet = spreadsheet;
    }
    
    /** Returns the whole Cell object located at position (row, col).<P>
     * Similar to getValueAt except the casting to a Cell object is handled before returning.
     * @param aRow The row number of the cell. Should be between 0 and the number of rows of the current spreadsheet.
     * @param aColumn The column number of the cell. Should be between 0 and the number of columns of the current spreadsheet.
     * @return  Returns the UMCell object at position (row, col). <P>Returns null 
     * if out of bounds. If a cell object does not exist there, it creates an empty cell and returns that. */
    public UMCell getCellAt(int aRow, int aColumn) {
        /* check for out of bounds */
        if (aRow < 0 || aRow >= getRowCount() ||
        aColumn < 0 || aColumn >= getColumnCount()) {
            return null;
        }
        
        UMCell temp = (UMCell)super.getValueAt(aRow, aColumn);
        return temp;
    }
    
    
   /** Returns the Object located at position (row, col).<P>
     * Similar to getCellAt except the object is not casted to a Cell before returning.<P>
     * It is implemented because TableModel requires this method return an Object.
     * @param aRow The row number of the cell. Should be between 0 and the number of rows of the current spreadsheet.
     * @param aColumn The column number of the cell. Should be between 0 and the number of columns of the current spreadsheet.
     * @return  Returns the Object at position (row, col). <P>Returns null 
     * if out of bounds. If a cell object does not exist there, it creates an empty cell and returns that. */
    public Object getValueAt(int aRow, int aColumn) {
        return getCellAt(aRow, aColumn);
    }
    
    /** Sets the value of the cell by simply passing its parameters to doSetValueAt. 
     * <P> This function is called by JTable automatically, which means a user has to manually input something.
     * @param aValue The data object to set the cell to.
     * @param aRow Row number of the cell. Should be between 0 and the number of rows of the current spreadsheet.
     * @param aColumn  Column number of the cell. Should be between 0 and the number of columns of the current spreadsheet.
     * @see doSetValueAt
     */     
    public void setValueAt(Object aValue, int aRow, int aColumn) {
        doSetValueAt(aValue, aRow, aColumn);
    }
    
    /** Sets the cell value to the passed object aValue. If aValue starts with a "=" then it will attempt to treat it as a formula.
     * Otherwise, the plain text value will be used.
     * @param aValue The data object to set the cell to.
     * @param aRow Row number of the cell. Should be between 0 and the number of rows of the current spreadsheet.
     * @param aColumn  Column number of the cell. Should be between 0 and the number of columns of the current spreadsheet. */    
    public void doSetValueAt(Object aValue, int aRow, int aColumn) {
        m_modified = true;
        if (aValue == null)
            aValue = new String("");
        if (aValue instanceof String) {
            String input = (String)aValue;
            // ADDED BY AL
            if (input.startsWith("'")) {
                Formula f = null;
                try {
                    f = new Formula(input, aRow, aColumn);
                    setCellAt(f, aRow, aColumn);
                    getCellAt(aRow, aColumn).setValue(input.substring(1));
                } catch (ParserException e) {
                    f = new Formula(input, aRow, aColumn, e);
                    setCellAt(f, aRow, aColumn);
                    getCellAt(aRow, aColumn).setValue(input.substring(1));
                }
                
                /* try making it a formula */
            } else if (input.startsWith("=")) {
                Formula form = null;
                System.out.println("Table Model: FORMULA:  " +input.substring(1));
                //create formula and its value and put in a cell
                try {
                    form = new Formula(input.substring(1), aRow, aColumn);
                    setCellAt(form, aRow, aColumn);

                } catch (ParserException e) {
                    // no parsing
                    form = new Formula(input.substring(1), aRow, aColumn, e);
                    setCellAt(form, aRow, aColumn);
                    getCellAt(aRow, aColumn).setValue(e);

                }
            } else {
                try{
                    Float data = new Float(input);
                    setCellAt(data, aRow, aColumn);
                }catch (NumberFormatException e2) {
                    /* all else fails treat as string */
                    setCellAt(aValue, aRow, aColumn);
                }
                
            }
        } else {
            /* it is an object, assume it is exactly
             * what should the cell be set to
             */
            if (aValue instanceof Formula) {
                try {
                    Formula form = (Formula)aValue;
                    setCellAt(new Formula(form, aRow, aColumn), aRow, aColumn);
                } catch (ParserException e) {
                    //	    errorMessage(e.toString());
                    Formula form2 = (Formula)aValue;
                    form2 = new Formula(form2.toString(), aRow, aColumn, e);
                    setCellAt(form2, aRow, aColumn);
                    getCellAt(aRow, aColumn).setValue(e);
                }
            } else {
                setCellAt(aValue, aRow, aColumn);
            }
        }
        
        /* it is an object, assume it is exactly
         * what should the cell be set to
         */
        // ADD this later to see if it is already of good object like formula
        
    }
    /** Determines if a cell is editable by making sure the column of the cell does not equal 0.
     * @param row The row number of the cell.
     * @param column The column number of the cell.
     * @return  Returns true if column != 0. False otherwise.*/    
    
    public boolean isCellEditable(int row, int column) {
        return column != 0;
    }
    
    /** JTable uses this method to determine the default renderer and editor for each cell.<P>
     * Right now it simply returns UMCell.class, which tells JTable to use UMCellRenderer and UMCellEditor.
     * @param c Column number to determine the class of. Currently the column number does not matter since only cell objects exist in TableModel. This parameter may be used later when different objects exist.
     * @return  Returns UMCell.class */    
    public Class getColumnClass(int c) {
        /* only cell objects in this TableModel */
        return UMCell.class;
    }
    
    
    /** Sets the modified state of the current document to the passed parameter.
     * @param modified  True sets state to modified. False unmodified*/
    public void setModified(boolean modified) {
        this.m_modified = modified;
        // enable/disable the "Save" button
    }
    
    /** Gets the current modified state of the document.
     * @return  Returns true if modified. False otherwise. */  
    public boolean isModified() {
        return m_modified;
    }
    
    /** Returns the spreadsheet table associated with the current UMTableModel.
     * @return  Returns the current spreadsheet table. */ 
    public JTable getTable() {
        return m_spreadsheet.getTable();
    }
    
    /** Determines if the cell at position (row, col) is empty.
     * @param row Row number of the cell. Should be between 0 and the number of rows of the current spreadsheet.
     * @param col Column number of the cell. Should be between 0 and the number of columns of the current spreadsheet.
     * @return  Returns true if the cell is empty. False otherwise. */    
    public boolean isEmptyCell(int row, int col) {
        return getCellAt(row, col).getValue().equals("");
    }
    
    /** Translates the given column number to the appropriate column letter.
     * @param col Column number. Should be greater than the base column of the current spreadsheet.
     * @return  String containing the letter of the column. Returns "" if col is greater than the base column of the current spreadsheet.*/    
    public String getColumnName(int col) {
        //Debug.println("name of column "+col);
        if (col < TerpSpreadSheet.m_baseCol)
            return "";
        else {
            String str = String.valueOf(Node.translateColumn(col));
            // Debug.println(Node.translateColumn(str) + " " +
            //                   str + "" + col);
            return str;
        }
    }
    
     /** Gets the current range of cells that are selected on the current spreadsheet.
      * @return  Returns the CellRange that is selected. */    
    public CellRange getSelection() {
        int ux = m_spreadsheet.m_baseCol ,
        uy = m_spreadsheet.m_baseRow ,
        lx = m_spreadsheet.m_baseCol,
        ly = m_spreadsheet.m_baseRow;
        JTable table = m_spreadsheet.getTable();
        
        if (table.getSelectedColumn() != -1) {
            ux  = table.getSelectedColumn();
            lx  = ux + table.getSelectedColumnCount() - 1;
        }
        if (table.getSelectedRow() != -1) {
            uy = table.getSelectedRow();
            ly = uy + table.getSelectedRowCount() - 1;
        }
        
        
        return new CellRange(uy,ly, ux,lx, this);
    }

    /** Fills the range of cells with null.
     * @param range Range of cells to be cleared. */      
    public void clearRange(CellRange range) {
        fill(range,null);
    }

    /** Calls doSetValueAt to set the value of each cell in the range to the data object.
     * @param range Cell Range to be filled. 
     * @param data Data object to fill cells with. 
     * @see doSetValueAt
     */     
    protected void fill(CellRange range, Object data){
        for(int row=range.getStartRow(); row<=range.getEndRow();row++) {
            for(int col=range.getStartCol();col<= range.getEndCol();col++) {
                doSetValueAt(data,row,col);
            }
        }
    }
    
    /**
 * Calculates the numeric value of the cell at position (row, col) and returns it.
     * @param row Row number of the cell. Should be between the base row and the number of rows of the current spreadsheet.
     * @param col Column number of the cell. Should be between the base column and the number of columns of the current spreadsheet.
     * @return  Numeric value of the cell. 
     * @throws ParserException Throws if the cell at position (row, col) is null.
     */        
    public Number getNumericValueAt(int row, int col) throws ParserException {
        UMCell cell = getCellAt(row, col);
        if (cell != null) {
            int type = cell.getType();
            if (type == UMCell.FORMULA) {
                Object value = cell.getValue();
                Formula form = cell.getFormula();
                // if need recalc
                if (form.needsRecalc()) {
                    try {
                        value = form.evaluate(this, row, col);
                        cell.setValue(value);
                    }
                    catch (ParserException e) {
                        cell.setValue(e);
                        value = e;
                    }
                }
                
                if (value instanceof ParserException)
                    throw (ParserException)value;
                else
                    return (Number)cell.getValue();
            }
            else if (type == UMCell.NUMBER)
                return (Number)cell.getValue();
            else
                return new Float(0);
        }
        else
            // a string or null
            //	return new Float(0);
            throw new ParserException("#REFS?");
        
    }
    
    /**
     * Determines if a cell at these coordinates is a formula cell
     *
     * @param aRow Row number of the cell. Should be between the base row and the number of rows of the current spreadsheet.
     * @param aColumn Column number of the cell. Should be between the base column and the number of columns of the current spreadsheet.
     * @return True only if the cell at those coordinates is a formula. False otherwise.
     */
    public boolean isFormula(int aRow, int aColumn) {
        UMCell temp = getCellAt(aRow, aColumn);
        return ((temp != null) && (temp.getType() == UMCell.FORMULA));
    }
    
    /** Called by setValueat. Assumes that the object passed to it is a correct object.
     * @param input Data object to set the cell value to.
     * @param aRow Row number of the destination cell. Should be between the base row and the number of rows of the current spreadsheet.
     * @param aColumn Column number of the destination cell. Should be between the base column and the number of columns of the current spreadsheet. 
     *
     * @see UMCell#setFormula
     * @see UMCell#setData
     */
    public void setCellAt(Object input, int aRow, int aColumn) {

        UMCell temp = getCellAt(aRow, aColumn);
        
        /* if for some reason value out of bounds ignore */
        if (temp != null) {

            //always remove references old formula referred to
            removeRefs(aRow, aColumn);
            
            //insert new formula
            if (input instanceof Formula) {

                temp.setFormula((Formula)input);
                if (isLoop(new CellPoint(aRow, aColumn))) {
                    ParserException loop = new ParserException("#LOOP?");
                    Formula form2 = new Formula(input.toString(), aRow, aColumn, loop);
                    setCellAt(form2, aRow, aColumn);
                    getCellAt(aRow, aColumn).setValue(loop);
                    updateRefs(aRow, aColumn);
                    return;
                }else {
                    addRefs(aRow, aColumn);
                    System.out.println(" calling CHECKING THE FORMULAR ");
                    recalculate(aRow, aColumn);
                    updateRefs(aRow, aColumn);
                    
                }
            } else {
              //treat as normal data cell
                temp.setData(input);
                updateRefs(aRow, aColumn);
            }
        }
    }
    
    /**
     * This method should be called with a cell that is set as a formula
     * cell (although it does nothing if it is not a formula). When a formula
     * is entered, it may reference other cells. These cells need to be
     * notified that if they are changed, to notify this formula cell.
     * This method adds the cell coordinates to the reference list of each
     * cell that the formula references.
     *
     * @param aRow Row number of the cell. Should be between the base row and the number of rows of the current spreadsheet.
     * @param aColumn Column number of the cell. Should be between the base column and the number of columns of the current spreadsheet.
     */
    public void addRefs(int aRow, int aColumn) {
        if (isFormula(aRow, aColumn)) {
            Formula temp = getCellAt(aRow, aColumn).getFormula();
            TreeSet list = temp.getDependency();
            //Debug.println("inside UMTablemodel of add refs"+ aRow +' '+ aColumn);
            /* use formula's dependency to find cells
             * that need to notify it if their values
             * change
             */
            Iterator it = list.iterator();
            CellPoint thisRef = new CellPoint(aRow, aColumn);
            while (it.hasNext()) {
                CellPoint update = (CellPoint)it.next();
                UMCell field = getCellAt(update.getRow(), update.getCol());
                
                //test of cell found was out of bounds
                if (field != null)
                    field.addRef(thisRef);
            }
        }
    }
    
    /**
     * This method removes this formula cell from the reference lists of all
     * cells it references. If this is not a formula cell, it does nothing.
     * This method should be called with a formula cell is being changed to a
     * non-Formula cell.
     *
     * @param aRow Row number of the cell to remove from reference list
     * @param aColumn Column number of the cell to remove from reference list
     */
    public void removeRefs(int aRow, int aColumn) {
        if (isFormula(aRow, aColumn)) {
            Formula temp = getCellAt(aRow, aColumn).getFormula();
            TreeSet list = temp.getDependency();
            
            /* use formula dependcy list to go to cells that it references
             * then remove its entry form their reference list
             */
            Iterator it = list.iterator();
            CellPoint thisRef = new CellPoint(aRow, aColumn);
            while (it.hasNext()) {
                CellPoint update = (CellPoint)it.next();
                UMCell field = getCellAt(update.getRow(), update.getCol());
                
                if (field != null)
                    field.removeRef(thisRef);
                
            }
        }
    }
    
    /**
     * This method updates the values of all cells that
     * reference the cell at position (aRow, aColumn). It recursively updates all cells that depend on this
     * one, and the cells that depend on those, and so on.
     *
     * @param aRow Row number of the cell to update. Should be between the base row and the number of rows of the current spreadsheet.
     * @param aColumn Column number of the cell to update. Should be between the base column and the number of columns of the current spreadsheet.
     */
    public void updateRefs(int aRow, int aColumn) {
        
        UMCell temp = getCellAt(aRow, aColumn);
        if (temp == null)
            return;
        
        TreeSet set = getRefs(aRow, aColumn);
        Debug.println("refs set for "+Node.translateColumn(aColumn)+
        Node.translateRow(aRow)+": "+set);
        
        // mark it as "needsRecalc";
        Iterator it = set.iterator();
        while (it.hasNext()) {
            CellPoint point = (CellPoint)it.next();
            Formula formula = getCellAt(point.getRow(), point.getCol()).getFormula();
            
            formula.setNeedsRecalc(true);
            // make sure JTable refreshes it
            //fireTableCellUpdated(point.getRow(), point.getCol());
        }
        
        // recalculate
        it = set.iterator();
        while (it.hasNext()) {
            CellPoint point = (CellPoint)it.next();
            try {
                getNumericValueAt(point.getRow(), point.getCol());
            }
            catch (ParserException e) {
            }
            
            // make sure JTable refreshes it
            fireTableCellUpdated(point.getRow(), point.getCol());
        }
        
        //make sure to tell JTable things have changed
        fireTableCellUpdated(aRow, aColumn);
    }
    
	/**
	   * This method gets the set of cells that will be affects
	   * by a value change for the cpecified cell.
	   *
	   * @param row the row
	   * @param col the column
	   *
	   * @return TreeSet -returns the set thats affected
	   */
    private TreeSet getRefs(int row, int col) {
        TreeSet set = new TreeSet();
        getRefs(row, col, set);
        return set;
    }
    
    /**
     * This method is a helper method for getReds(int, int).
     * It recursively gets refs for each cell and merges it into
     * the set.
     *
     * @param row the row
     * @param col the column
     * @param set the current of cells
     *
     */
    private void getRefs(int row, int col, TreeSet set) {
        UMCell cell = getCellAt(row, col);
        if (cell == null || !cell.hasRefs())
            return;
        
        Iterator it = cell.getRefs().iterator();
        while (it.hasNext()) {
            CellPoint point = (CellPoint)it.next();
            set.add(point);
            getRefs(point.getRow(), point.getCol(), set);
        }
    }
    //Billy added the function here
    /**
     * This method recalculates all cells in the table
     *
     * @see #insertRow
     * @see #insertColumn
     * @see #removeRow
     * @see #removeColumn
     *
     */
    public void recalculateAll() {
        for (int i = 1; i < getRowCount(); i++)
            for (int j = 1; j < getColumnCount(); j++) {
                addRefs(i, j);
                recalculate(i, j);
            }
    }
	/**
   * Starting from cell, detect potential reference loops.
   *
   * @param cell specified CellPoint
   * @return boolean -Returns True/Flase if there is a loop
   */
    private boolean isLoop(CellPoint cell) {
        return isLoop(cell, new TreeSet());
    }
    
	/**
	  * Starting from cell, detect potential reference loops.
	  *
	  * @param cell specified CellPoint
	  * @param set the set of all the cells in the tree
	  * @return boolean -Returns True/Flase if there is a loop
	  */
    private boolean isLoop(CellPoint cell, TreeSet set) {
        if (set.contains(cell))
            return true;
        
        UMCell objCell = getCellAt(cell.getRow(), cell.getCol());
        if (objCell == null)
            return false;
        Formula formula = objCell.getFormula();
        if (formula == null)
            return false;
        
        set.add(cell);
        
        Iterator it = formula.getDependency().iterator();
        while (it.hasNext()) {
            CellPoint newCell = (CellPoint)it.next();
            boolean ret = isLoop(newCell, set);
            if (ret)
                return true;
        }
        
        set.remove(cell);
        return false;
    }
    
	/**
 * This is a static method that parses string input passed to it from somewhere else. <P>
	 * Tries to create a formula, then a number. Returns the original string if all else fails.
	 * @param input String to parse
	 * @param c Cell to place parsed string in.
	 * @return Object - Returned value
	 */
    private static Object fieldParser(String input, CellPoint c) {
        if (input == null)
            return new String("");
        
        int row = c.getRow();
        int col = c.getCol();
        
        
        /* try making it a formula */
        if (input.startsWith("=")) {
            Formula form = null;
            
            //create formula and its value and put in a cell
            try {
                return new Formula(input.substring(1), row, col);
            }catch (ParserException e) {
                // no parsing
                return new Formula(input.substring(1), row, col, e);
            }
            
        }else {
            /* try making it a number */
            try {
                return new Float(input);
                
            }catch (NumberFormatException e2) {
                /* all else fails treat as string */
                return input;
            }
        }
    }
    
    /** This method does not recognize formula strings. It is used for dialogue
     * box input where there will be no formulas inputted or expected to be
     * inputted.
     * @param input input string to parse
     * @return appropriate object after parsing
     */
    
    public static Object fieldParser(String input) {
        if (input == null)
            return new String("");
        
        /* try making it a number */
        try {
            return new Float(input);
        }catch (NumberFormatException e) {
            /* all else fails treat as string */
            return input;
        }
    }
    
    /**
     * This is a wraper method of the Formula class's evaluate method. This
     * method recalculates the value of a formula at the given coordinates.
     * If the coordinates do not specify a formula it does nothing.
     *
     * @param aRow the row coordinate
     * @param aColumn the column coordinate
     */
    
    public void recalculate(int aRow, int aColumn) {
        if (isFormula(aRow, aColumn)) {
            
            try {
                System.out.println("Recalc: " );
                Number eVal = Formula.evaluate(this, aRow, aColumn);
                
                //we set the value here
                //Debug.println(aRow +"***"+ aColumn +"***");
                getCellAt(aRow, aColumn).setValue(eVal);
                //Debug.println("CHECKING THE FORMULAR "+ eVal);
            }
            catch (ParserException e) {
                //set value as the appropriate error message
                getCellAt(aRow, aColumn).setValue(e);
            }
        }
    }
    /**
     * This is the version of recalculate that takes a CellPoint object as an
     * argument.
     *
     * @param x the coordinates of the cell to be updated
     */
    public void recalculate(CellPoint x) {
        recalculate(x.getRow(), x.getCol());
    }
    
    /** Another version of to string.
     *
     * @return string representation
     */    
    public String toString() {
        String text = toString(new CellRange(TerpSpreadSheet.m_baseRow, getRowCount()-1,
        TerpSpreadSheet.m_baseCol, getColumnCount()-1), false);
        return text;
    }
    
     /** Converts the given range of cells into a string by concatenating their 
      * string values from upper, left-most to lower, right-most with the characters 
      * \t between columns and \n between rows.
      * @param range The range of cells in the table to be converted. Must be a valid range in the current spreadsheet.
      * @param byValue True to return the actual value or formula. False to return the displayed value (calculated formula).
      * @return  String representing the given range of cells. 
      * @see Formula#fixRelAddr
      * @see FileOp#saveTableModel
      * @see EditOp#cut
      * @see EditOp#copy
      * @see Clipboard
      */
    public String toString(CellRange range, boolean byValue) {
        StringBuffer sbf=new StringBuffer();
        
        for (int i=range.getStartRow(); i<=range.getEndRow(); i++) {
            for (int j=range.getStartCol(); j<=range.getEndCol(); j++) {
                if (byValue) {
                    sbf.append(getValueAt(i, j));
                }
                else {
                    UMCell cell = getCellAt(i, j);
                    // if cell is not empty
                    if(cell != null)
												//modified by Xun
                        //sbf.append(cell.toString());
                        sbf.append(cell.CellToString());
                }
                
                if (j<range.getEndCol())
                    sbf.append("\t");
            }
            sbf.append("\n");
        }
        String text = sbf.toString();
        return text;
    }
    
    /** Converts a string to values in a range of cells. For example, multiple 
     * lines of text will be pasted into multiple rows.
     * @param text The string to convert.
     * @param rowOff The row offset (currently not used).
     * @param colOff The column offset (currently not used).
     * @param range The range of cells to paste the string into.
     * @see Formula#fixRelAddr
     * @see FileOp#openTableModel
     * @see EditOp#paste
     * @see Clipboard
     */ 
    public void fromString(String text, int rowOff, int colOff, CellRange range) {
        
        try {
            BufferedReader in = new BufferedReader(new StringReader(text));
            String line;
            int row = range.getStartRow();
            
            while (row <= range.getEndRow()) {
                line = in.readLine();
                //System.out.println("------>>>>> " + line);
                
                int index;
                int prev = 0;
                
                // set col to startCol before each loop
                int col = range.getStartCol();
                String value;
                
                while (col <= range.getEndCol()) {
                    index = line.indexOf('\t', prev);
                    if (index >= 0) {
                        value = line.substring(prev, index);
                    }
                    else {
                        value = line.substring(prev);
                    }
                    
                    if (value.startsWith("=")) {
                        // need to fix relative address
                        /*value =
                            Formula.fixRelAddr(value.substring(1),
                                               rowOff, colOff);
                        if (value == null)
                            value = new String("=$REFS$0");
                        else
                            value = "="+value;*/
                    }
                    
                    doSetValueAt(value, row, col);
                    //System.out.println(""+row+','+col+": "+value);
                    
                    prev = index+1;
                    // increment column number
                    col++;
                    
                    if (index == -1)
                        break;
                }
                
                row++;
            }
        }
        catch (Exception e) {
        }
    }

    /** toString is used to convert a range of cells into a string.
     * One row per line, and each column is tab-delimited.
     * Different from toString in that it appends code that tells font properties.
     * @param range Range of cells in the table.
     * @param byValue True gets the value. False gets the formula.
     * @return  The inputed string. 
     * @see Formula#fixRelAddr
     * @see FileOp#saveTableModel
     * @see EditOp#cut
     * @see EditOp#copy
     * @see Clipboard
     **/    
    public String toString1(CellRange range, boolean byValue) {
        //System.out.println("To String has been called");
        StringBuffer sbf=new StringBuffer();
        String cpy;
        int fontsize, indx, fontstate;
        int bGround = getTable().getBackground().getRGB(),
        fGround = getTable().getForeground().getRGB(),
        iTop = 0,
        iBot = 0,
        iLeft = 0,
        iRight = 0;
        for (int i=range.getStartRow(); i<=range.getEndRow(); i++) {
            for (int j=range.getStartCol(); j<=range.getEndCol(); j++) {
                if (byValue) {
                    sbf.append(getCellAt(i, j).getValue().toString());
                }
                else {
                    UMCell cell = getCellAt(i, j);
                    // if cell is not empty
                    if(cell != null){
												//modified by Xun
                        //sbf.append(cell.toString());
                        sbf.append(cell.CellToString());
                        try{
                            if(cell.getFont()== null){
                                //System.out.print("Cell Get font is null Initialising Inddex");
                                indx=0;
                                fontsize=12;
                                fontstate=0;
                            }
                            else{
                                indx=EditMenuOp.getFontIndex(cell.getFont().getName());
                                fontsize=cell.getFont().getSize();
                                if(cell.getFont().isPlain())
                                    fontstate=0;
                                else if(cell.getFont().isItalic())
                                    fontstate=1;
                                else
                                    fontstate=2;
                            }
                            Color bColor = cell.getBackground();
                            Color fColor = cell.getForeground();
                            LinesBorder border = (LinesBorder) cell.getBorder();
                            bGround = getTable().getBackground().getRGB();
                            fGround = getTable().getForeground().getRGB();
                            iTop = 0;
                            iBot = 0;
                            iLeft = 0;
                            iRight = 0;
                            
                            if (bColor != null)
                                bGround = bColor.getRGB();
                            if (fColor != null)
                                fGround = fColor.getRGB();
                            if (border != null) {
                                //System.out.println("Border is not null");
                                Insets ins = border.getBorderInsets(null);
                                iTop = ins.top;
                                iBot = ins.bottom;
                                iRight = ins.right;
                                iLeft = ins.left;
                            }
                            
                            sbf.append("\\*"+indx+"*"+fontsize+"*"+fontstate+"*"+
                            bGround+"*" + fGround + "*" + iTop +
                            "*" + iBot + "*" + iRight + "*" +
                            iLeft + "*/");
                            
                        } catch (Exception e){
                            System.out.print("Error Occurance");
                            e.printStackTrace();
                        }
                    }
                }
                
                if (j<range.getEndCol()){
                    sbf.append("\t");
                }
            }
            sbf.append("\n");
        }
        String text = sbf.toString();
        
        return text;
    }

    /** Converts a string to values in a range of cells.
     * One row per line, and each column is tab-delimited.
     * @param text The inputed string.
     * @param rowOff The row offset.
     * @param colOff The column offset.
     * @param range  The range to paste. 
     * @see Formula#fixRelAddr
     * @see FileOp#openTableModel
     * @see EditOp#paste
     * @see Clipboard
     */      
    public void fromString1(String text, int rowOff, int colOff, CellRange range) {
        boolean camein = false;
        try {
            BufferedReader in = new BufferedReader(new StringReader(text));
            String line;
            int row = range.getStartRow();
            while (row <= range.getEndRow()) {
                line = in.readLine();
                int index;
                int prev = 0, prev2=0;
                
                // set col to startCol before each loop
                int col = range.getStartCol(),loc;
                String value, temp, start1, fontproperty = null;
                start1= "\\*";
                while (col <= range.getEndCol()) {
                    camein = false;
                    index = line.indexOf('\t', prev);
                    if (index >= 0) {
                        //value= line.substring(prev, index);
                        temp = line.substring(prev,index);
                        //System.out.println("Temp: "+temp);
                        //WILL ALWAYS FIND START1
                        loc = temp.indexOf(start1, prev2);
                        if(loc >= 0) {
                            value = temp.substring(0,loc);
                            fontproperty = temp.substring(loc);
                        } else
                            value = temp;
                    } else {
                        temp = line.substring(prev);
                        //System.out.println("Temp: "+temp+" Code 3 Index"+index);
                        //WILL ALWAYS FIND START1
                        loc = temp.indexOf(start1, prev2);
                        if(loc >= 0) {
                            value = temp.substring(0,loc);
                            fontproperty = temp.substring(loc);
                        } else
                            value = temp;
                    }
                    if (value.startsWith("PS")) {
                        System.out.println("Printing special");
                        setValueAt(value.substring(2), row, col);
                        Number numeric = getNumericValueAt(row, col);
                        System.out.println("NUMERIC MAN "+ numeric);
                        setValueAt(numeric, row, col);
                        camein = true;	
                    } else if (value.startsWith("'")) { 
                        System.out.println("Direct cut/copy & paste operation.  Eg: '=SQRT(9) => =SQRT(9)");
                        setCellAt(value, row, col);
                        getCellAt(row, col).setData(value.substring(1));
                        setFontproperty(fontproperty, getCellAt(row, col));
                        camein = true;
                    } else if (value.startsWith("=")) {
                        // need to fix relative address
                        //System.out.println("Temp: "+temp);
                        String ret = Formula.fixRelAddr(value.substring(1), rowOff, colOff);
                        
                        if (value == null)
                            value = new String("=$REFS$0");
                        else 
                            value = "="+ret;
                        Formula form = null;
                        
                        //create formula and its value and put in a cell
                        
                        try {
                            form = new Formula(value.substring(1),row, col);
                            setCellAt(form, row,col);
                            camein = true;
                            System.out.println("7");
                            //set font properties
                            setFontproperty(fontproperty, getCellAt(row, col));
                        }
                        catch (ParserException e) {
                            System.out.println("8");
                            // no parsing
                            form = new Formula(value.substring(1), row,col, e);
                            setCellAt(form, row,col);
                            camein = true;
                            getCellAt(row,col).setValue(e);
                            setFontproperty(fontproperty, getCellAt(row, col));
                        }
                    }
                    System.out.println("9");
                    if(!camein)
                        setCellAt(value,row, col,fontproperty);
                    prev = index+1;
                    // increment column number
                    col++;
                    if (index == -1)
                        break;
                }
                row++;
            }
        }
        catch (Exception e) {
        }
    }
    
    /** Similar copy of SetCellAt
     * @param input Data object to set cell value to.
     * @param aRow Row number of the cell.
     * @param aColumn Column number of the cell.
     * @param fontdata  String containing font information.
     */  
    public void setCellAt(Object input, int aRow, int aColumn, String fontdata) {
        //System.out.println("Set cell At UMCell Called");
        UMCell temp = getCellAt(aRow, aColumn);
        
        /* if for some reason value out of bounds ignore */
        if (temp != null) {
            //always remove references old formula referred to
            removeRefs(aRow, aColumn);
            //insert new formula
            if (input instanceof Formula) {
                temp.setFormula((Formula) input);
                
                if (isLoop(new CellPoint(aRow, aColumn))) {
                    ParserException loop = new ParserException("#LOOP?");
                    Formula form2 = new Formula(input.toString(), aRow, aColumn, loop);
                    setCellAt(form2, aRow, aColumn);
                    getCellAt(aRow, aColumn).setValue(loop);
                    updateRefs(aRow, aColumn);
                    return;
                } else {
                    
                    addRefs(aRow, aColumn);
                    recalculate(aRow, aColumn);
                    updateRefs(aRow, aColumn);
                }
            } else {
            /* Default font values
                if (fontdata != null)
                {
                int fname=0, size=12, state=0,
                    bGround = getTable().getBackground().getRGB(),
                    fGround = getTable().getForeground().getRGB(),
                    iTop= 0,
                    iBot = 0,
                    iRight = 0,
                    iLeft = 0;
                String indx,fontsize, fontstate,
                    bColorStr, fColorStr,
                    iTopStr, iBotStr, iLeftStr,
                    iRightStr;
                String parseStr = fontdata.substring(2,fontdata.length());
                //System.out.println(parseStr);
                if(fontdata != null){
                  StringTokenizer tk = new StringTokenizer(parseStr, "*");
                  try {
                  indx= tk.nextToken();
                  fontsize= tk.nextToken();
                  fontstate = tk.nextToken();
                  bColorStr = tk.nextToken();
                  fColorStr = tk.nextToken();
                  iTopStr = tk.nextToken();
                  iBotStr = tk.nextToken();
                  iRightStr = tk.nextToken();
                  iLeftStr = tk.nextToken();
             
             
                    fname=Integer.parseInt(indx);
                    size=Integer.parseInt(fontsize);
                    state=Integer.parseInt(fontstate);
                    bGround = Integer.parseInt(bColorStr);
                    fGround = Integer.parseInt(fColorStr);
                    iTop = Integer.parseInt(iTopStr);
                    iBot = Integer.parseInt(iBotStr);
                    iRight = Integer.parseInt(iRightStr);
                    iLeft = Integer.parseInt(iLeftStr);
                    } catch (Exception e){
                      System.out.println("Format Error in SetCellAt");
                      input = new String(input.toString()) + fontdata;
                  }
                }
             
                Font a;
                if(state==0)
                  a=new Font(Fontlist[fname], Font.PLAIN,size);
                else if(state==1)
                  a=new Font(Fontlist[fname], Font.ITALIC,size);
                else
                  a=new Font(Fontlist[fname], Font.BOLD,size);
                temp.setFont(a);
             
                Color bColor = new Color(bGround);
                Color fColor = new Color(fGround);
                LinesBorder border = null;
                if (iTop == 2 ||
                    iBot == 2 ||
                    iRight == 2 ||
                    iLeft == 2)
               {
                   Insets ins = new Insets(1,1,1,1);
                   ins.top = iTop;
                   ins.bottom = iBot;
                   ins.left = iLeft;
                   ins.right = iRight;
                   border = new LinesBorder(Color.black, ins);
               }
                temp.setBackground(bColor);
                temp.setForeground(fColor);
                temp.setBorder(border);
                }*/
                
                //FONT ACTIVITIES
                try{
                    setFontproperty(fontdata, temp);
                } catch( Exception e){
                    System.out.println("Format Error in SetCellAt");
                    input = new String(input.toString()) + fontdata;
                }
                //temp.copy(input);
                if (input == null)
                    input = new String("");
                if (input instanceof String) {
                    try {
                        String hope = (String)input;
                        Float data = new Float(hope);
                        temp.setData(data);
                        temp.setFormula(null);
                    } catch (NumberFormatException e){
                        //string and formulas
                        temp.setData(input);
                    }
                }
                
                updateRefs(aRow, aColumn);
                
            }
            
        }
    }
    
	 /**
* Sets the font property of the passed cell.
	* @param fontdata String containing font information.
	* @param temp -The cell
	* @throws java.lang.Exception - If the font data is bad
	*/      
    private void setFontproperty(String fontdata, UMCell temp) throws Exception{
        if (fontdata != null){
            int fname=0, size=12, state=0, bGround = getTable().getBackground().getRGB(),
            fGround = getTable().getForeground().getRGB(),
            iTop= 0,
            iBot = 0,
            iRight = 0,
            iLeft = 0;
            String indx,fontsize, fontstate,bColorStr, fColorStr,iTopStr, iBotStr, iLeftStr,iRightStr;
            String parseStr = fontdata.substring(2,fontdata.length());
            //System.out.println(parseStr);
            if(fontdata != null){
                StringTokenizer tk = new StringTokenizer(parseStr, "*");
                try {
                    indx= tk.nextToken();
                    fontsize= tk.nextToken();
                    fontstate = tk.nextToken();
                    bColorStr = tk.nextToken();
                    fColorStr = tk.nextToken();
                    iTopStr = tk.nextToken();
                    iBotStr = tk.nextToken();
                    iRightStr = tk.nextToken();
                    iLeftStr = tk.nextToken();
                    
                    fname=Integer.parseInt(indx);
                    size=Integer.parseInt(fontsize);
                    state=Integer.parseInt(fontstate);
                    bGround = Integer.parseInt(bColorStr);
                    fGround = Integer.parseInt(fColorStr);
                    iTop = Integer.parseInt(iTopStr);
                    iBot = Integer.parseInt(iBotStr);
                    iRight = Integer.parseInt(iRightStr);
                    iLeft = Integer.parseInt(iLeftStr);
                }catch (Exception e){
                    throw new Exception("Font Property Error");
                }
            }
            
            Font a;
            if(state==0)
                a=new Font(Fontlist[fname], Font.PLAIN,size);
            else if(state==1)
                a=new Font(Fontlist[fname], Font.ITALIC,size);
            else
                a=new Font(Fontlist[fname], Font.BOLD,size);
            temp.setFont(a);
            
            Color bColor = new Color(bGround);
            Color fColor = new Color(fGround);
            LinesBorder border = null;
            if (iTop == 2 ||iBot == 2 ||iRight == 2 ||iLeft == 2){
                Insets ins = new Insets(1,1,1,1);
                ins.top = iTop;
                ins.bottom = iBot;
                ins.left = iLeft;
                ins.right = iRight;
                border = new LinesBorder(Color.black, ins);
            }
            temp.setBackground(bColor);
            temp.setForeground(fColor);
            temp.setBorder(border);
        }
        
    }
    /** Another version of to string.
     *
     * @return the string representation of a cell
     */    
    public String ConvertToString() {
        String text = CelltoString(new CellRange(TerpSpreadSheet.m_baseRow,
        getRowCount()-1,
        TerpSpreadSheet.m_baseCol,
        getColumnCount()-1), false);
        
        
        return text;
    }

    /** Converts a range of cells into a string.
     * One row per line, and each column is tab-delimited.
     * @param range Range of cells to convert
     * @param byValue True to return the actual value or formula. False to return the cells (including font formats) in String format.
     * @return  String representing the range of cells. 
     * @see Formula#fixRelAddr
     * @see FileOp#saveTableModel
     * @see EditOp#cut
     * @see EditOp#copy
     * @see Clipboard
     */    
    public String CelltoString(CellRange range, boolean byValue) {
        StringBuffer sbf=new StringBuffer();
        //System.out.println("IN CELL TO STRING");
        for (int i=range.getStartRow(); i<=range.getEndRow(); i++) {
            for (int j=range.getStartCol(); j<=range.getEndCol(); j++) {
                if (byValue) {
                    sbf.append(getValueAt(i, j));
                }
                else {
                    UMCell cell = getCellAt(i, j);
                    // if cell is not empty
                    if(cell != null)
												//modified by Xun
                        //sbf.append(cell.toString());
                        sbf.append(cell.CellToString());
                    //-------------------------------
                    // following was added by JJdESi4
                    else
                        sbf.append(" ");
                    //-------------------------------
                }
                
                if (j<range.getEndCol())
                    sbf.append("\t");
            }
            sbf.append("\n");
        }
        String text = sbf.toString();
        //System.out.println(text);
        //System.out.println("======================================");
        return text;
    }
    /** Converts a string containing font data to values in a range of cells.
     * One row per line, and each column is tab-delimited.
     * @param text The string containing font data to place in the range.
     * @param rowOff Row offset (not used).
     * @param colOff Column offset (not used).
     * @param range Range of cells to paste to. 
     * @see doSetCellValueAt
     * @see Formula#fixRelAddr
     * @see FileOp#openTableModel
     * @see EditOp#paste
     * @see Clipboard
     */  
    public void CellfromString(String text, int rowOff, int colOff, CellRange range) {
        
        try {
            BufferedReader in = new BufferedReader(new StringReader(text));
            String line;
            int row = range.getStartRow();
            
            while (row <= range.getEndRow()) {
                line = in.readLine();
                //System.out.println("------>>>>> " + line);
                
                int index;
                int prev = 0;
                
                // set col to startCol before each loop
                int col = range.getStartCol();
                String value;
                
                while (col <= range.getEndCol()) {
                    index = line.indexOf('\t', prev);
                    if (index >= 0) {
                        value = line.substring(prev, index);
                    }
                    else {
                        value = line.substring(prev);
                    }
                    
                    //System.out.println(""+row+','+col+": "+value);
                    doSetCellValueAt(value, row, col);
                    //System.out.println(""+row+','+col+": "+value);
                    
                    prev = index+1;
                    // increment column number
                    col++;
                    
                    if (index == -1)
                        break;
                }
                
                row++;
            }
            for (int i=range.getStartRow(); i<=range.getEndRow(); i++) {
                for (int j=range.getStartCol(); j<=range.getEndCol(); j++) {
                    updateRefs(i, j);
                }
            }
        }
        catch (Exception e) {
        }
    }
    
    /** Sets the value of the cell at position (aRow, aColumn) to aValue.
     * @param aValue String containing font data to set cell value to.
     * @param aRow Row number of the cell. Should be between the base row and the number of rows of the current spreadsheet.
     * @param aColumn Column number of the cell.
     *                Should be between the base column and
     *                the number of columns of the current spreadsheet.
     */    
    
    public void doSetCellValueAt(String aValue, int aRow, int aColumn) {
        //System.out.println("############ "+aValue);
        m_modified = true;
        UMCell newCell = new UMCell();
        String temp = new String();
        StringTokenizer strToken = new StringTokenizer((String)aValue, "/");
        Color bColor = null;
        Color fColor = null;
        Font font = null;
        Object value = null;
        Formula expression = null;
        UMCell tempCell = getCellAt(aRow, aColumn);
        
        //parse out the background color
        temp = strToken.nextToken();
        //System.out.println("Back Color ===> " + temp);
        if (temp.compareTo("null") != 0) {
            StringTokenizer intToken = new StringTokenizer(temp, "-");
            int blue = (Integer.valueOf(intToken.nextToken())).intValue();
            int red = (Integer.valueOf(intToken.nextToken())).intValue();
            int green = (Integer.valueOf(intToken.nextToken())).intValue();
            //System.out.println(blue + " = " + red + " = " + green + " = ");
            bColor = new Color(red, green, blue);
            //newCell.setBackground(bColor);
            tempCell.setBackground(bColor);
            //System.out.println("======= >>> " + newCell.getBackground());
        }
        
        //parse out the foreground color
        temp = strToken.nextToken();
        //System.out.println("Fore Color ===> " + temp);
        if (temp .compareTo("null") != 0) {
            StringTokenizer intToken = new StringTokenizer(temp, "-");
            int blue = (Integer.valueOf(intToken.nextToken())).intValue();
            int red = (Integer.valueOf(intToken.nextToken())).intValue();
            int green = (Integer.valueOf(intToken.nextToken())).intValue();
            //System.out.println(blue + " = " + red + " = " + green + " = ");
            fColor = new Color(red, green, blue);
            //newCell.setForeground(fColor);
            tempCell.setForeground(fColor);
        }
        
        //parse out the font style
        temp = strToken.nextToken();
        //System.out.println("Font ===> "+temp);
        if (temp.compareTo("null") != 0) {
            StringTokenizer fontToken = new StringTokenizer(temp, "-");
            String name = fontToken.nextToken();
            int size = (Integer.valueOf(fontToken.nextToken())).intValue();
            int style = (Integer.valueOf(fontToken.nextToken())).intValue();
            //System.out.println(name + " = " + size + " = " + style + " = ");
            font = new Font(name, style, size);
            //newCell.setFont(font);
            tempCell.setFont(font);
            //font = Font.getFont(temp);
            //System.out.println("Font ===> "+font);
        }
        
        temp = strToken.nextToken();
        if (temp.compareTo("null") != 0) {
            StringTokenizer borderToken = new StringTokenizer(temp, "-");
            int top = (Integer.valueOf(borderToken.nextToken())).intValue();
            int bottom = (Integer.valueOf(borderToken.nextToken())).intValue();
            int left = (Integer.valueOf(borderToken.nextToken())).intValue();
            int right = (Integer.valueOf(borderToken.nextToken())).intValue();
            Insets ins = new Insets(top, left, bottom, right);
            Border theBorder = new LinesBorder
            (UIManager.getColor("Table.gridColor"), ins);
            //newCell.setBorder(theBorder);
            tempCell.setBorder(theBorder);
        }
        //setValueAt(newCell, aRow, aColumn);
        if (strToken.hasMoreElements()) {
            temp = strToken.nextToken();
            doSetValueAt(temp, aRow, aColumn);
        }
    }
    
    
    // serey add
        /**
 * This searches of an object starting from a cell point.
         * @param begin cellpoint to begin search
         * @param input input
         * @param matchCase true if case sensitive (only active if goal is a string)
         * @param matchCell true if goal must equal entire content of cell (only
         *    active if goal is a string)
         * @return the CellPoint of next occurence of goal
         */
    public CellPoint look(CellPoint begin, Object input, boolean matchCase,
			     boolean matchCell) {
        int startRow = begin.getRow();
        int startCol = begin.getCol();
        int i;
        int j;
        
        String goal = input.toString();
        String test = null;
        
        Pattern p = Pattern.compile(goal);
        Pattern pUpper = Pattern.compile(goal.toUpperCase()); // for case insensitive searching
        Matcher m = null;

        
       System.out.println("looking for " + goal);
       System.out.println("starting at " + startRow + ", " + startCol);
       
       // start searching from the current cell
       boolean starting = true;
       for (i = startRow; i < getRowCount()-1; i++){
            for (j = 1; j < getColumnCount()-1; j++){
                if (starting){
                    j = startCol + 1;
                    starting = false;
                }
                
                test = getCellAt(i, j).getValue().toString();
                
                if (matchCase && !matchCell){
                    m = p.matcher(test);
                    if (m.find()){
                        return new CellPoint(i, j);
                    }
                } else if (matchCase && matchCell){
                    if (test.equals(goal)){
                        return new CellPoint(i, j);
                    }
                } else if (!matchCase && matchCell){
                    if (test.equalsIgnoreCase(goal)){
                        return new CellPoint(i, j);
                    }
                } else if (!matchCase && !matchCell){
                    String testUpper = test.toUpperCase();
                    m = pUpper.matcher(testUpper);
                    if (m.find()){
                        System.out.println("match found at " + i + ", " + j);
                        return new CellPoint(i, j);
                    }
                }
            }
        }
       
       // if inputValue hasn't been found by now, start over from cell A1 up until the original start cell
       for (i = 0; i <= startRow; i++){
            for (j = 1; j < getColumnCount()-1; j++){
                
	if ((i != startRow)||(j != startCol))
                    return null;
                
                test = getCellAt(i, j).getValue().toString();
                
                if (matchCase && !matchCell){/*FAULT:	if (!matchCase || matchCell){*/
                    m = p.matcher(test);
                    if (m.find()){
                        return new CellPoint(i, j);
                    }
                } else if (matchCase && matchCell){/*FAULT:	} else if (!matchCase || !matchCell){*/
                    if (test.equals(goal)){
                        return new CellPoint(i, j);
                    }
                } else if (!matchCase && matchCell){/*FAULT:	} else if (matchCase || !matchCell){*/
                    if (test.equalsIgnoreCase(goal)){
                        return new CellPoint(i, j);
                    }
                } else if (!matchCase && !matchCell){/*FAULT:	} else if (matchCase || matchCell){*/
                    String testUpper = test.toUpperCase();
                    m = pUpper.matcher(testUpper);
                    if (m.find()){
                        System.out.println("match found at " + i + ", " + j);
                        return new CellPoint(i, j);
                    }
                }
            }
        }
       
        System.out.println("no match");
        return null;
    }
    
    
    
    /**
     * check whether the deletion is safe
     *
     * @param range the range to delete
     * @param byRow whether it's deletion by row
     * @return true if it's safe
     */
    public boolean isDeletionSafe(CellRange range, boolean byRow) {
	int rowOff, colOff;
	CellRange needCheck;
	
	if (byRow) {
	    rowOff = -range.getHeight();
	    colOff = 0;
	    if (range.getEndRow() == getRowCount()-1)
		return true;
	    
	    needCheck = new CellRange(range.getEndRow()+1, getRowCount()-1,
				      TerpSpreadSheet.m_baseCol, getColumnCount()-1);
	}
	else {
	    rowOff = 0;
	    colOff = -range.getWidth();
	    if (range.getEndCol() == getColumnCount()-1)
		return true;
	    needCheck = new CellRange(TerpSpreadSheet.m_baseRow, getRowCount()-1,
				      range.getEndCol()+1, getColumnCount()-1);
	}

	for (int i = needCheck.getStartRow(); i <= needCheck.getEndRow(); i++)
	    for (int j = needCheck.getStartCol(); j <= needCheck.getEndCol();
		 j++) {
		UMCell cell = getCellAt(i, j);
		if (cell.isFormula() &&
		    !Formula.isSafe(cell.getFormula(), rowOff, colOff)) {
		    Debug.println("relative addresses become invalid");
		    return false;
		}
	    }

	return true;
	
    }
    
    
    /** This method removes rows in the range.
     * @param deletionRange CellRange that contains the rows to delete
     */
    public void removeRow(CellRange deletionRange) {
        /* since the insertion point is given by a selected cell
         * there will never be an out of bounds error
         */
        clearRange(deletionRange);
        
        /* first row to delete */
        int row = deletionRange.getStartRow();
        
        /* number of rows to delete including the first */
        int removeNum = deletionRange.getHeight();
        
        //coordinates of last cell in spreadsheet
        int lastRow = getRowCount() - 1;
        int lastCol = getColumnCount() - 1;
        
        //everything lower than rows to remove must be copied to be shifted
        CellRange range = new CellRange(row + removeNum, lastRow,
					TerpSpreadSheet.m_baseCol, lastCol);
        CopyPaste scrap = new CopyPaste(this, range, true);
           
        for(int i = 0; i < removeNum; i++) {
            super.removeRow(getRowCount() - 1);
        }
        
        
        //shift relevent rows up
        scrap.paste(this, new CellPoint(row, TerpSpreadSheet.m_baseCol));    

	recalculateAll();
	
	JTable table = m_spreadsheet.getTable();
	// set selection
	if (table.getSelectedColumnCount() == 0) {
	    setSelection(new CellRange(row, row, TerpSpreadSheet.m_baseCol, TerpSpreadSheet.m_baseCol));
	}
    }
    
    
    
    /** This method removes columns in the range.
     * @param deletionRange the range that contains the columns to delete
     */
    public void removeColumn(CellRange deletionRange) {

        /* since the insertion point is given by a selected cell
         * there will never be an out of bounds error
         */
        
        /* first column to delete */
        int col = deletionRange.getStartCol();
        
        /* number of columns to delete including col */
        int removeNum = deletionRange.getWidth();
        
        //last entry of table
        int lastRow = getRowCount() - 1;
        int lastCol = getColumnCount() - 1;
        
        
        /* everything to the right of the columns to remove
         * need to be copied to be shifted right
         */
        CellRange range = new CellRange(TerpSpreadSheet.m_baseRow, lastRow,
                                        col + removeNum, lastCol);
	
        
        CopyPaste scrap = new CopyPaste(this, range, true);
        
        JTable table = m_spreadsheet.getTable();
        for(int i = 0; i < removeNum; i++) {
            // delete old column
            removeColumn();
	    TableColumnModel tm = table.getColumnModel();
	    tm.removeColumn(tm.getColumn(tm.getColumnCount() - 1));
	}
        
        //shift clipboard elements right
        scrap.paste(this, new CellPoint(TerpSpreadSheet.m_baseRow, col));

	//	updateRefs(refs);
	recalculateAll();
	
	// set selection
	if (table.getSelectedColumnCount() == 0) {
	    setSelection(new CellRange(TerpSpreadSheet.m_baseRow, TerpSpreadSheet.m_baseRow,
				       col, col));
	}

        //	fireTableStructureChanged();
	m_spreadsheet.setBaseColumnWidth();
    }
    
	/** This method removes columns in the range. Called by Public removeColumn.
	 *	see #removeColumn
	 **/
    private void removeColumn() {

        int lastRow = getRowCount() - 1;
        int lastCol = getColumnCount() - 1;
        
        //remove the data from cells to delete to maintain references
        clearRange(new CellRange(TerpSpreadSheet.m_baseRow, lastRow, lastCol, lastCol));
        
        Iterator it = dataVector.iterator();
        while (it.hasNext()) {
            /* Since deleting B makes C the
             * new B, the reference lists in cells of old "B"  
             * should not change. So, we only shift the data in cells right
             * and deleted the last columns.
             */
            Vector temp = (Vector)it.next();
            temp.removeElementAt(getColumnCount() - 1);
        }
        
        //update inherited field from  DefaultTableModel
        columnIdentifiers.removeElementAt(columnIdentifiers.size() - 1);
        
        // Notification is generated within the GUI
    }
    
    
    /** This inserts a new row in the spreadsheet.
     *
     * @param insertRange the rand to insert.
     */    
    public void insertRow(CellRange insertRange) {
        /* since the insertion point is given by a selected cell
         * there will never be an out of bounds error
         */
        
        /* insert starting at this coordinate */
        int row = insertRange.getStartRow();
        
        /* number of rows to insert including row */
        int insertNum = insertRange.getHeight();
        
        //coordinates of last cell of table
        int lastRow = getRowCount() - 1;
        int lastCol = getColumnCount() - 1;
        
        /* copy things below these new rows
         * The max is to prevent inserts above the column headers
         */
        CellRange range = new CellRange(Math.max(row, TerpSpreadSheet.m_baseRow), lastRow, TerpSpreadSheet.m_baseCol, lastCol);
        CopyPaste scrap = new CopyPaste(this, range, true);
        
        //add the rows to the end
        for(int i = 0; i < insertNum; i++) {
            addRow();
        }
        
        //shift old rows down
        scrap.paste(this, new CellPoint(row + insertNum, TerpSpreadSheet.m_baseCol));
        
        recalculateAll();
        
        JTable table = m_spreadsheet.getTable();
        // set selection
        if (table.getSelectedColumnCount() == 0) {
            setSelection(new CellRange(row, row, TerpSpreadSheet.m_baseCol, TerpSpreadSheet.m_baseCol));
        }
        
    }
    
    
    
    /**
     *  Adds row to end of table
     */
    private void addRow() {
        
        //create a new row with appropriate label
        Vector rowData = new Vector();
        rowData.add(0, new UMCell(new Integer(getRowCount() + 1)));
        
        //add it to the table
        super.addRow(rowData);
        
        for(int i = 1; i < getColumnCount(); i++) {
            super.setValueAt(new UMCell(""), getRowCount() - 1, i);
        }
    }
    
    
    /**
     * set table selection to the range sel
     *
     * @param sel the range to be selected
     */
    public void setSelection(CellRange sel) {
        JTable table = m_spreadsheet.getTable();
        // validate sel
        int maxRow = table.getRowCount()-1;
        int maxCol = table.getColumnCount()-1;
        
        int startRow = sel.getStartRow();
        int startCol = sel.getStartCol();
        int endRow = sel.getEndRow();
        int endCol = sel.getEndCol();
        
        table.setColumnSelectionInterval(Math.min(startCol, maxCol),
        Math.min(endCol, maxCol));
        table.setRowSelectionInterval(Math.min(startRow, maxRow),
        Math.min(endRow, maxRow));
        
    }
    
    
    
    /** This method inserts columns to the left of the range with the number of
     * new columns equal to the number of columns in the range.
     * @param insertRange range of cells to add new columns to the left of
     * creates the same number of new columns as range has
     */
    
    public void insertColumn(CellRange insertRange) {
        /* since the insertion point is given by a selected cell
         * there will never be an out of bounds error
         */
        Debug.println("insertRange: "+insertRange);
        /* start inserting at this coordinate */
        int col = insertRange.getStartCol();
        
        /* number of columns to insert including col */
        int insertNum = insertRange.getWidth();
        
        //the coordinates of the last cell in table
        int lastRow = getRowCount() - 1;
        int lastCol = getColumnCount() - 1;
        
        /* everything right to new columns must be shifted right
         * so cut them to the clipboard. So if col is "C" then it
         * is also copied. The max is a guard for inserting
         * before the label column.
         */
        CellRange range = new CellRange(TerpSpreadSheet.m_baseRow, lastRow, Math.max(col, TerpSpreadSheet.m_baseCol), lastCol);
        CopyPaste scrap = new CopyPaste(this, range, true);
        
        JTable table = m_spreadsheet.getTable();
        TableColumnModel tm = table.getColumnModel();
        TableColumn column = tm.getColumn(col);
        //add the new columns to the end
        for(int i = 0; i < insertNum; i++) {
            int curCol = lastCol+i+1;
            addColumn();
            TableColumn newcol = new TableColumn(curCol,
            column.getPreferredWidth());
            //	    TableColumn column = tm.getColumn(tm.getColumnCount() - 1);
            //            TableColumn newcol = new TableColumn(tm.getColumnCount(),
            //                                                 column.getPreferredWidth());
            newcol.setHeaderValue(Node.translateColumn(curCol));
            tm.addColumn(newcol);
        }
        
        //shift relevant columns left
        scrap.paste(this, new CellPoint(TerpSpreadSheet.m_baseRow, col + insertNum));
        
        recalculateAll();
        // set selection
        
        if (table.getSelectedColumnCount() == 0) {
            setSelection(new CellRange(TerpSpreadSheet.m_baseRow, TerpSpreadSheet.m_baseRow,
            col, col));
        }
        
        //	sharp.setBaseColumnWidth();
        //fireTableStructureChanged();
        m_spreadsheet.setBaseColumnWidth();
    }
    
    
    
    /** Helper method for insertColumn. This method will not
     * send the appropriate notification to JTable. Please use insertColumn method
     * instead.
     */
    private void addColumn() {
        
        columnIdentifiers.addElement(null);

        /* Initialize the new column */
        Iterator it = dataVector.iterator();
        
        //Give column the appropriate label 
        if (it.hasNext()) {
            UMCell temp = new UMCell(Node.translateColumn(getColumnCount() - 1));
            ((Vector)it.next()).addElement(temp);
        }
        
        //initialize cells
        while (it.hasNext()) {
            ((Vector)it.next()).addElement(new UMCell(""));
        }

        // Generate notification
	/*        newColumnsAdded(new TableModelEvent(this, 0, getRowCount() - 1,
		  getColumnCount() - 1, TableModelEvent.INSERT));
	*/
    }    
}
